/* Copyright (c) 1992, 1998, 2000 John E. Davis
 * This file is part of JED editor library source.
 *
 * You may distribute this file under the terms the GNU General Public
 * License.  See the file COPYING for more information.
 */
    
/* For undo, these flags assumed to be <= 0xFF */
#define CDELETE  0x1
#define CINSERT  0x2
#define LDELETE  0x4
#define NLINSERT  0x8   /* causes marks to be moved to next line */
#define NLDELETE  0x10    /* opposite of above */
  
extern void jed_update_marks(int, int);  
extern int No_Screen_Update;
extern int jed_deln (int);
extern int jed_generic_deln (int);
extern int jed_del(void);
extern int jed_ins( char);
extern void insert_buffer(Buffer *);
extern int jed_quick_insert(register unsigned char *, int);
extern int jed_ins_chars(unsigned char *, int);
extern int jed_ins_char_n_times(char, int);
extern int jed_del_newline(void);
extern int jed_replace_char(unsigned char);
extern int jed_insert_string (char *s);
extern int jed_insert_newline (void);
extern int Suspend_Screen_Update;
extern int jed_prepare_for_modification (void);
